#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 13:02

import time
from . import _compat, MIXPANEL_EVENTS, MIXPANEL_PEOPLE
from .consumer import MixpanelConsumer, MixpanelBufferedConsumer
from .utils import dispatch
from tornado import gen


class ClientError(Exception):
    pass


class AsyncMixpanelClient(object):
    VERSION = '1.0.0'
    VERSION_NAME = 'python/tornado'

    def __init__(self, token, buffered=False, **settings):
        if not isinstance(token, _compat.string_types) or not len(token):
            raise ClientError('Token not found')
        self._token = token
        if not buffered:
            self._consumer = MixpanelConsumer(**settings)
        else:
            self._consumer = MixpanelBufferedConsumer(**settings)

    @property
    def consumer(self):
        return self._consumer

    def _sanitize_argument(self, value, default):
        return default if value is None else value

    @gen.coroutine
    def import_data(self, api_key, distinct_id, event_name, timestamp,
                    properties=None, meta=None, callback=None):
        prop = {
            'token': self._token,
            'distinct_id': distinct_id,
            'time': int(timestamp),
            'mp_lib': self.VERSION_NAME,
            '$lib_version': self.VERSION,
        }
        prop.update(self._sanitize_argument(properties, {}))
        data = {'event': event_name, 'properties': prop}
        data.update(self._sanitize_argument(meta, {}))
        response = yield self._consumer.send(MIXPANEL_EVENTS, data,
                                             api_key=api_key)
        dispatch(response, callback)

    @gen.coroutine
    def track(self, distinct_id, event_name, properties=None, meta=None,
              callback=None):
        prop = {
            'token': self._token,
            'distinct_id': distinct_id,
            'time': int(time.time()),
            'mp_lib': self.VERSION_NAME,
            '$lib_version': self.VERSION,
        }
        prop.update(self._sanitize_argument(properties, {}))
        data = {'event': event_name, 'properties': prop}
        data.update(self._sanitize_argument(meta, {}))
        response = yield self._consumer.send(MIXPANEL_EVENTS, data)
        dispatch(response, callback)

    @gen.coroutine
    def alias(self, alias_id, original, meta=None, callback=None):
        data = {
            'event': '$create_alias',
            'properties': {
                'distinct_id': original,
                'alias': alias_id,
                'token': self._token
            }
        }
        data.update(self._sanitize_argument(meta, {}))
        response = yield self._consumer.fetch(MIXPANEL_EVENTS, data)
        dispatch(response, callback)

    @gen.coroutine
    def people(self, distinct_id, message, meta=None, callback=None):
        data = {
            '$token': self._token,
            '$distinct_id': distinct_id,
            '$time': int(time.time() * 1000)
        }
        data.update(message, **self._sanitize_argument(meta, {}))
        response = yield self._consumer.send(MIXPANEL_PEOPLE, data)
        dispatch(response, callback)

    def people_set(self, distinct_id, properties, meta=None, callback=None):
        return self.people(distinct_id, {'$set': properties}, meta, callback)

    def people_set_once(self, distinct_id, properties, meta=None,
                        callback=None):
        return self.people(distinct_id, {'$set_once': properties}, meta,
                           callback)

    def people_add(self, distinct_id, properties, meta=None, callback=None):
        return self.people(distinct_id, {'$add': properties}, meta, callback)

    def people_append(self, distinct_id, properties, meta=None, callback=None):
        return self.people(distinct_id, {'$append': properties}, meta, callback)

    def people_union(self, distinct_id, properties, meta=None, callback=None):
        return self.people(distinct_id, {'$union': properties}, meta, callback)

    def people_unset(self, distinct_id, properties, meta=None, callback=None):
        return self.people(distinct_id, {'$unset': properties}, meta, callback)

    def people_delete(self, distinct_id, meta=None, callback=None):
        return self.people(distinct_id, {'$delete': ''}, meta, callback)

    def people_track_charge(self, distinct_id, amount, properties=None,
                            meta=None, callback=None):
        properties = self._sanitize_argument(properties, {})
        properties['$amount'] = amount
        return self.people_append(distinct_id, {'$transactions': properties},
                                  meta, callback)

    def people_clear_charges(self, distinct_id, meta=None, callback=None):
        return self.people_unset(distinct_id, ["$transactions"], meta, callback)



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 13:57

import time
import traceback
from tornado import gen, ioloop
from tornado_mixpanel.client import AsyncMixpanelClient


@gen.coroutine
def run():
    client = AsyncMixpanelClient('<mixpanel-token>', True)
    raw_input('Press (enter) to continue...')

    try:
        username = int(time.time())
        print 'Tracking...'

        for i in xrange(10):
            r = yield client.track(username, 'item_%s' % i, {'i': i})
            print i, r == False
            time.sleep(1)
        print '-*-' * 20

        r = yield client.consumer.flush()
        print r

    except:
        print traceback.format_exc()

    ioloop.IOLoop.current().stop()


if __name__ == '__main__':
    run()
    ioloop.IOLoop.instance().start()


#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 12:48

# Version

version = '0.2'
version_tuple = (0, 2, 0, 0)


# Mixpanel Settings

MIXPANEL_EVENTS = 'events'
MIXPANEL_PEOPLE = 'people'
MIXPANEL_IMPORTS = 'imports'

MIXPANEL_ENDPOINTS = {
    MIXPANEL_EVENTS: 'https://api.mixpanel.com/track',
    MIXPANEL_PEOPLE: 'https://api.mixpanel.com/engage',
    MIXPANEL_IMPORTS: 'https://api.mixpanel.com/import'
}

MIXPANEL_ENDPOINTS_KEYS = frozenset(MIXPANEL_ENDPOINTS.keys())



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 13:07

# TODO(berna): Support for Python 3.0+

import datetime

try:
    import ujson
    json = ujson
except ImportError:
    try:
        import simplejson
        json = simplejson
    except ImportError:
        import json


bytes_types = str
string_types = basestring
unicode_types = unicode
list_types = (tuple, list)

primitive_types = \
    (complex, int, float, long, bool, str, basestring, unicode, tuple, list)

datetime_types = \
    (datetime.date, datetime.time, datetime.datetime, datetime.timedelta)



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 13:03

from . import _compat
from tornado import gen


# Decorators

def value_or_none(func):
    def decorator(value, *args, **kwargs):
        if value is None:
            return None
        return func(value, *args, **kwargs)
    return decorator


def string_or_error(func):
    def decorator(value, *args, **kwargs):
        if not isinstance(value, _compat.string_types):
            raise TypeError('Must be a string type.')
        return func(value, *args, **kwargs)
    return decorator


def dict_or_error(func):
    def decorator(value, *args, **kwargs):
        if not isinstance(value, dict):
            raise TypeError('Must be a dict type.')
        return func(value, *args, **kwargs)
    return decorator


# Strings

@string_or_error
def want_bytes(value, encoding='utf-8', errors='strict'):
    if isinstance(value, _compat.unicode_types):
        value = value.encode(encoding, errors)
    return value


# Dictionary

@dict_or_error
def setdefault(source, key, default):
    source[key] = source.get(key, None) or default


# Async

def dispatch(response, callback=None):
    if callback is None:
        raise gen.Return(response)
    callback(response)


# Misc

@value_or_none
def complex_types(value):
    if isinstance(value, _compat.unicode_types):
        return want_bytes(value)
    elif isinstance(value, _compat.primitive_types):
        return value
    elif isinstance(value, _compat.datetime_types):
        return value.isoformat()
    return _compat.bytes_types(value)


#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 13:57

import traceback
from tornado import gen, ioloop
from tornado_mixpanel.client import AsyncMixpanelClient


@gen.coroutine
def run():
    client = AsyncMixpanelClient('<mixpanel-token>')
    raw_input('Press (enter) to continue...')

    try:
        r = yield client.track(
            'user-xxxx', 'steps', {'step_one': True, 'step_two': False})
        print r

        r = yield client.people_set(
            'client-xxxx', {'fullname': 'Alejandro Bernardis'})
        print r

        r = yield client.people_append(
            'client-xxxx', {'age': 31, 'locale': 'es_AR'})
        print r

    except:
        print traceback.format_exc()

    ioloop.IOLoop.current().stop()


if __name__ == '__main__':
    run()
    ioloop.IOLoop.instance().start()


#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 03/Feb/2015 13:02

import base64
from . import _compat, MIXPANEL_ENDPOINTS_KEYS, \
    MIXPANEL_ENDPOINTS
from .utils import setdefault, complex_types, want_bytes, \
    dispatch
from collections import deque
from datetime import datetime, timedelta
from functools import wraps
from tornado import gen
from tornado.httpclient import HTTPRequest, HTTPResponse, AsyncHTTPClient, \
    HTTPError
from urllib import urlencode


class ConsumerError(Exception):
    pass


def is_endpoint(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        endpoint = kwargs.get('endpoint', args[0])
        if endpoint not in MIXPANEL_ENDPOINTS_KEYS:
            raise gen.Return(ConsumerError(
                'No such endpoints "%s". Valid endpoints are one of %s.'
                % (endpoint, MIXPANEL_ENDPOINTS_KEYS)))
        return method(self, *args, **kwargs)
    return wrapper


def is_buffered(method):
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not getattr(self, '_buffered', False):
            raise gen.Return(ConsumerError(
                'Async buffered consumer not supported'))
        return method(self, *args, **kwargs)
    return wrapper


class MixpanelConsumer(object):
    def __init__(self, endpoints=None, **settings):
        AsyncHTTPClient.configure('tornado.curl_httpclient.CurlAsyncHTTPClient')
        for item in settings.get('endpoints', endpoints or {}).keys():
            if item not in MIXPANEL_ENDPOINTS_KEYS:
                raise ConsumerError('Endpoint not supported: %s' % item)
        setdefault(settings, 'endpoints', MIXPANEL_ENDPOINTS)
        self._endpoints = settings['endpoints']

    def _sanitize_message(self, message):
        if isinstance(message, deque):
            message = list(message)
        return _compat.json\
            .dumps(message, default=complex_types, separators=(',', ':'))

    def _sanitize_body(self, message, api_key=None):
        if isinstance(message, (deque, list, tuple, dict)):
            message = self._sanitize_message(message)
        elif not isinstance(message, _compat.string_types):
            raise ConsumerError('Invalid message, must be a string')
        message = base64.b64encode(want_bytes(message))
        values = {'data': message, 'verbose': 1, 'ip': 0}
        if api_key is not None:
            values['api_key'] = api_key
        return urlencode(values, True).encode('utf-8')

    def _sanitize_headers(self, headers=None):
        return headers

    def _sanitize_response(self, response=None, error=None):
        if isinstance(error, HTTPError):
            response = error.response
        if not isinstance(response, HTTPResponse):
            raise ConsumerError('Response must be a '
                                '"tornado.httpclient.HTTPResponse".')
        body = _compat.json.loads(response.body)
        if response.code != 200 or response.error or body.get('status', 0) != 1:
            raise ConsumerError(body, response)
        return body

    @gen.coroutine
    @is_endpoint
    def fetch(self, endpoint, message, headers=None, **kwargs):
        try:
            request = HTTPRequest(self._endpoints[endpoint], 'POST')
            request.headers = self._sanitize_headers(headers)
            request.body = self._sanitize_body(message)
            for key, value in kwargs.items():
                setattr(request, key, value)
            try:
                response = yield AsyncHTTPClient().fetch(request)
                response = self._sanitize_response(response)
            except HTTPError, e:
                response = self._sanitize_response(error=e)
            except Exception, e:
                response = e
        except Exception, e:
            response = ConsumerError(e.message)
        raise gen.Return(response)

    @gen.coroutine
    @is_endpoint
    def send(self, endpoint, message, callback=None, **kwargs):
        try:
            response = yield self.fetch(endpoint, message, **kwargs)
        except ConsumerError, e:
            response = e
        except Exception, e:
            response = ConsumerError(None, e)
        dispatch(response, callback)


class MixpanelBufferedConsumer(MixpanelConsumer):
    ALL = 'all'
    ENDPOINT = 'endpoint'

    def __init__(self, endpoints=None, flush_after=60, flush_first=True,
                 max_size=50, **kwargs):
        self._current = {key: deque() for key in MIXPANEL_ENDPOINTS_KEYS}
        self._buffers = {key: deque() for key in MIXPANEL_ENDPOINTS_KEYS}
        self._flush_first = flush_first
        self._flush_after = timedelta(0, flush_after)
        self._last_flushed = None if not flush_after else datetime.utcnow()
        self._errors = None
        self._max_size = min(max_size, 50)
        super(MixpanelBufferedConsumer, self).__init__(endpoints, **kwargs)

    @property
    def errors(self):
        return self._errors

    def _should_flush(self, endpoint=None):
        full = endpoint and len(self._current[endpoint]) >= self._max_size
        stale = self._last_flushed is None
        if not stale and self._flush_after:
            stale = datetime.utcnow() - self._last_flushed > self._flush_after
        if stale:
            return self.ALL
        if full:
            return self.ENDPOINT
        return False

    @gen.coroutine
    def _prepare_flush(self, endpoint=None, **kwargs):
        errors = []
        endpoints = [endpoint] if endpoint in MIXPANEL_ENDPOINTS_KEYS \
            else MIXPANEL_ENDPOINTS_KEYS
        for endpoint in endpoints:
            current = self._current[endpoint]
            if not len(current):
                continue
            buffering = self._buffers[endpoint]
            buffering.extend(current)
            current.clear()
            try:
                # TODO(berna): collect and analyze results
                message = '[{0}]'.format(','.join(buffering))
                yield self.fetch(endpoint, message, **kwargs)
            except Exception, e:
                current.extendleft(buffering)
                errors.append(ConsumerError(endpoint, e))
            buffering.clear()
        response = len(errors) == 0
        self._errors = None if response else errors
        raise gen.Return(response)

    @gen.coroutine
    def flush(self, callback=None, **kwargs):
        response = yield self._prepare_flush(**kwargs)
        dispatch(response, callback)

    @gen.coroutine
    @is_endpoint
    def send(self, endpoint, message, callback=None, **kwargs):
        response = False
        current = self._current[endpoint]
        current.append(self._sanitize_message(message))
        try:
            should = self._should_flush(endpoint)
        except:
            should = False
        if should == self.ALL:
            response = yield self._prepare_flush(**kwargs)
        elif should == self.ENDPOINT:
            response = yield self._prepare_flush(endpoint, **kwargs)
        dispatch(response, callback)



#!/usr/bin/env python2.7
# -*- coding: utf-8 -*-
# Copyright (c) 2014 Asumi Kamikaze Inc.
# Licensed under the MIT License.
# Author: Alejandro M. Bernardis
# Email: alejandro (dot) bernardis (at) asumikamikaze (dot) com
# Created: 02/Feb/2015 14:01

from tornado_mixpanel import version
from setuptools import setup, find_packages

with open('README.rst', 'r') as f:
    readme = f.read()

setup(
    name='tornado-mixpanel',
    version=version,
    url='https://github.com/alejandrobernardis/tornado-mixpanel',
    license='MIT',
    author='Alejandro M. Bernardis',
    author_email='alejandro.m.bernardis@gmail.com',
    description='An async client for mixpanel.',
    long_description=readme,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Natural Language :: Spanish',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities'
    ],
    packages=find_packages(exclude=['docs', 'examples', 'scripts', 'templates',
                                    'tests', 'libs']),
    package_data={'': ['CHANGES', 'LICENSE', 'AUTHORS', 'README.rst']},
    package_dir={'tornado_mixpanel': 'tornado_mixpanel'},
    include_package_data=True,
    zip_safe=False,
    platforms=['Linux'],
    install_requires=[line.strip() for line in open('requirements.txt', 'r')]
)
